@@ -21,6 +21,7 @@ import (
 	"context"
 	"fmt"
 	"log/slog"
+	"runtime/debug"
 	"sync"
 	"sync/atomic"
 
@@ -42,6 +43,7 @@ import (
 // Returns a context for managing the server.
 func Serve(
 	ctx context.Context,
+	bufVersion string,
 	wktBucket storage.ReadBucket,
 	container appext.Container,
 	controller bufctl.Controller,
@@ -49,15 +51,25 @@ func Serve(
 	stream jsonrpc2.Stream,
 	queryExecutor *incremental.Executor,
 ) (jsonrpc2.Conn, error) {
+	// Prefer build info version if available
+	if buildInfo, ok := debug.ReadBuildInfo(); ok && buildInfo.Main.Version != "" {
+		bufVersion = buildInfo.Main.Version
+	}
+
+	logger := container.Logger()
+	logger = logger.With(slog.String("buf_version", bufVersion))
+	logger.Info("starting LSP server")
+
 	conn := jsonrpc2.NewConn(stream)
 	lsp := &lsp{
 		conn: conn,
 		client: protocol.ClientDispatcher(
-			&connWrapper{Conn: conn, logger: container.Logger()},
+			&connWrapper{Conn: conn, logger: logger},
 			zap.NewNop(), // The logging from protocol itself isn't very good, we've replaced it with connAdapter here.
 		),
 		container:     container,
-		logger:        container.Logger(),
+		logger:        logger,
+		bufVersion:    bufVersion,
 		controller:    controller,
 		wasmRuntime:   wasmRuntime,
 		wktBucket:     wktBucket,
@@ -93,6 +105,7 @@ type lsp struct {
 	container appext.Container
 
 	logger           *slog.Logger
+	bufVersion       string // buf version, set at server creation
 	controller       bufctl.Controller
 	wasmRuntime      wasm.Runtime
 	fileManager      *fileManager
