@@ -17,10 +17,8 @@
 package buflsp
 
 import (
-	"encoding/json"
-	"strings"
-
 	"github.com/bufbuild/protocompile/experimental/report"
+	"github.com/bufbuild/protocompile/experimental/report/tags"
 	"github.com/bufbuild/protocompile/experimental/source/length"
 	"go.lsp.dev/protocol"
 )
@@ -30,14 +28,6 @@ import (
 // https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#position
 const positionalEncoding = length.UTF16
 
-// diagnosticData is a structure to hold the [report.Diagnostic] notes, help, and debug
-// messages, to marshal into JSON for the [protocol.Diagnostic].Data field.
-type diagnosticData struct {
-	Notes string `json:"notes,omitempty"`
-	Help  string `json:"help,omitempty"`
-	Debug string `json:"debug,omitempty"`
-}
-
 // reportLevelToDiagnosticSeverity is a mapping of [report.Level] to [protocol.DiagnosticSeverity].
 var reportLevelToDiagnosticSeverity = map[report.Level]protocol.DiagnosticSeverity{
 	report.ICE:     protocol.DiagnosticSeverityError,
@@ -50,7 +40,7 @@ var reportLevelToDiagnosticSeverity = map[report.Level]protocol.DiagnosticSeveri
 // corresponding [protocol.Diagnostic].
 func reportDiagnosticToProtocolDiagnostic(
 	reportDiagnostic report.Diagnostic,
-) (protocol.Diagnostic, error) {
+) protocol.Diagnostic {
 	diagnostic := protocol.Diagnostic{
 		Source:   serverName,
 		Severity: reportLevelToDiagnosticSeverity[reportDiagnostic.Level()],
@@ -70,20 +60,15 @@ func reportDiagnosticToProtocolDiagnostic(
 			},
 		}
 	}
-	data := diagnosticData{
-		Notes: strings.Join(reportDiagnostic.Notes(), "\n"),
-		Help:  strings.Join(reportDiagnostic.Help(), "\n"),
-		Debug: strings.Join(reportDiagnostic.Debug(), "\n"),
+	switch reportDiagnostic.Tag() {
+	case tags.UnusedImport:
+		diagnostic.Tags = []protocol.DiagnosticTag{
+			protocol.DiagnosticTagUnnecessary,
+		}
+	case tags.Deprecated:
+		diagnostic.Tags = []protocol.DiagnosticTag{
+			protocol.DiagnosticTagDeprecated,
+		}
 	}
-	bytes, err := json.Marshal(data)
-	if err != nil {
-		return protocol.Diagnostic{}, err
-	}
-	if bytes != nil {
-		// We serialize the bytes into a string before providing the structure to diagnostic.Data
-		// because diagnostic.Data is an interface{}, which is treated as a JSON "any", which
-		// will not cleanly deserialize.
-		diagnostic.Data = string(bytes)
-	}
-	return diagnostic, nil
+	return diagnostic
 }
