@@ -27,6 +27,7 @@ import (
 	"strings"
 	"unicode"
 
+	"github.com/bufbuild/buf/private/bufpkg/bufconnect"
 	"github.com/bufbuild/buf/private/bufpkg/bufmodule"
 	"github.com/bufbuild/protocompile/experimental/ast"
 	"github.com/bufbuild/protocompile/experimental/ast/predeclared"
@@ -189,6 +190,61 @@ func (s *symbol) References(includeDeclaration bool) []protocol.Location {
 	return references
 }
 
+// DocumentHighlights returns document highlights for the symbol within the current file.
+// This includes the definition (if in the same file) and all references in the same file.
+// All highlights use the [protocol.DocumentHighlightKindText] kind.
+func (s *symbol) DocumentHighlights() []protocol.DocumentHighlight {
+	// Don't highlight static symbols (services, methods, enum values)
+	if _, ok := s.kind.(*static); ok {
+		return nil
+	}
+
+	// Get the referenceable kind to find all references
+	referenceableKind, ok := s.kind.(*referenceable)
+	if !ok && s.def != nil {
+		// If the symbol isn't referenceable itself, but has a referenceable definition, use the
+		// definition for the references.
+		referenceableKind, ok = s.def.kind.(*referenceable)
+	}
+	if !ok {
+		return nil
+	}
+
+	// Don't highlight field names. Field names have referenceable kind directly on the symbol,
+	// whereas field type references have reference kind and reference a referenceable definition.
+	// Both have ir.SymbolKindField, so we distinguish by checking if s.kind is referenceable.
+	if _, isRefKind := s.kind.(*referenceable); isRefKind && s.ir.Kind() == ir.SymbolKindField {
+		return nil
+	}
+
+	var highlights []protocol.DocumentHighlight
+	// Add all references in the same file
+	for _, reference := range referenceableKind.references {
+		if reference.file.uri == s.file.uri {
+			highlights = append(highlights, protocol.DocumentHighlight{
+				Range: reference.Range(),
+				Kind:  protocol.DocumentHighlightKindText,
+			})
+		}
+	}
+
+	// Add the definition if it's in the same file
+	if s.def != nil && s.def.file.uri == s.file.uri {
+		highlights = append(highlights, protocol.DocumentHighlight{
+			Range: s.def.Range(),
+			Kind:  protocol.DocumentHighlightKindText,
+		})
+	} else if s.def == nil {
+		// If there's no separate definition, the symbol itself is the definition
+		highlights = append(highlights, protocol.DocumentHighlight{
+			Range: s.Range(),
+			Kind:  protocol.DocumentHighlightKindText,
+		})
+	}
+
+	return highlights
+}
+
 // LogValue provides the log value for a symbol.
 func (s *symbol) LogValue() slog.Value {
 	if s == nil {
@@ -234,12 +290,18 @@ func (s *symbol) FormatDocs() string {
 		builtin, _ := s.kind.(*builtin)
 		comments, ok := builtinDocs[builtin.predeclared.String()]
 		if ok {
+			// Use specific anchor for map, generic anchor for other builtins
+			anchor := "field-types"
+			if builtin.predeclared.String() == "map" {
+				anchor = "maps"
+			}
 			comments = append(
 				comments,
 				"",
 				fmt.Sprintf(
-					"`%s` is a Protobuf builtin. [Learn more on protobuf.com.](https://protobuf.com/docs/language-spec#field-types)",
+					"`%s` is a Protobuf builtin. [Learn more on protobuf.com.](https://protobuf.com/docs/language-spec#%s)",
 					builtin.predeclared,
+					anchor,
 				),
 			)
 			return strings.Join(comments, "\n")
@@ -302,8 +364,10 @@ func (s *symbol) GetSymbolInformation() protocol.SymbolInformation {
 		Kind:          kind,
 		Location:      location,
 		ContainerName: containerName,
-		// TODO: Use Tags with a protocol.CompletionItemTagDeprecated if the client supports tags.
-		Deprecated: isDeprecated,
+		Deprecated:    isDeprecated,
+		Tags: []protocol.SymbolTag{
+			protocol.SymbolTagDeprecated,
+		},
 	}
 }
 
@@ -358,11 +422,48 @@ func renameChangesForReferenceableSymbol(s *symbol, newName string) (map[protoco
 			NewText: newName,
 		}},
 	}
-	if def, ok := s.def.kind.(*referenceable); ok {
-		for _, reference := range def.references {
+	// Get the referenceable kind to find all references
+	referenceableKind, ok := s.kind.(*referenceable)
+	if !ok && s.def != nil {
+		// If the symbol isn't referenceable itself, but has a referenceable definition, use the
+		// definition for the references.
+		referenceableKind, ok = s.def.kind.(*referenceable)
+	}
+	if ok {
+		for _, reference := range referenceableKind.references {
+			newText := newName
+			// For option references (extension usages), preserve package qualification and parentheses.
+			// e.g., if renaming "(subpkg.testing)" to "validated", result should be "(subpkg.validated)"
+			if _, isOption := reference.kind.(*option); isOption {
+				spanText := reference.span.Text()
+				// Extract components: prefix (opening paren + package), suffix (closing paren)
+				prefix := ""
+				suffix := ""
+				nameOnly := spanText
+
+				// Check for opening parenthesis
+				if strings.HasPrefix(spanText, "(") {
+					prefix = "("
+					nameOnly = nameOnly[1:]
+				}
+				// Check for closing parenthesis
+				if strings.HasSuffix(nameOnly, ")") {
+					suffix = ")"
+					nameOnly = nameOnly[:len(nameOnly)-1]
+				}
+				// Check if there's package qualification (contains a dot)
+				if lastDot := strings.LastIndex(nameOnly, "."); lastDot != -1 {
+					// Preserve the package qualification
+					packageQualification := nameOnly[:lastDot+1]
+					newText = prefix + packageQualification + newName + suffix
+				} else {
+					// No package qualification, just preserve parens
+					newText = prefix + newName + suffix
+				}
+			}
 			changes[reference.file.uri] = append(changes[reference.file.uri], protocol.TextEdit{
 				Range:   reportSpanToProtocolRange(reference.span),
-				NewText: newName,
+				NewText: newText,
 			})
 		}
 	} else {
@@ -456,22 +557,28 @@ func (s *symbol) getDocsFromComments() string {
 	// traversing backwards for leading comments only.
 	tok, _ := def.Context().Stream().Around(def.Span().Start)
 	cursor := token.NewCursorAt(tok)
-	var seenNewline bool
+	// Count consecutive newlines. If we accumulate 2+ without encountering a comment,
+	// there's a blank line separating the comments from the symbol.
+	newlinesSeen := 0
+	if tok.Kind() == token.Space {
+		newlinesSeen = strings.Count(tok.Text(), "\n")
+		if newlinesSeen >= 2 {
+			return ""
+		}
+	}
 	for {
 		t := cursor.PrevSkippable()
 		if t.Kind() == token.Comment {
-			text := commentToMarkdown(t.Text())
-			if seenNewline {
-				text += "\n"
+			if isTrailingComment(t) {
+				break
 			}
-			seenNewline = false
+			text := commentToMarkdown(t.Text()) + "\n"
+			newlinesSeen = 0
 			comments = append(comments, text)
 		} else if t.Kind() == token.Space {
-			if strings.Contains(t.Text(), "\n") {
-				if seenNewline {
-					break
-				}
-				seenNewline = true
+			newlinesSeen += strings.Count(t.Text(), "\n")
+			if newlinesSeen >= 2 {
+				break
 			}
 		} else {
 			break
@@ -499,29 +606,46 @@ func (s *symbol) getDocsFromComments() string {
 			hasAnchor = isExtension
 		}
 
-		var bsrHost, bsrAnchor string
+		var module bufmodule.Module
+		var bsrHost string
 		if s.def.file.IsWKT() {
-			bsrHost = "buf.build/protocolbuffers/wellknowntypes"
+			bsrHost = bufconnect.DefaultRemote + "/protocolbuffers/wellknowntypes"
 		} else if fileInfo, ok := s.def.file.objectInfo.(bufmodule.FileInfo); ok {
-			bsrHost = fileInfo.Module().FullName().String()
+			module = fileInfo.Module()
+			bsrHost = module.FullName().String()
 		}
+
 		defFullName := s.def.ir.FullName()
 		if !hasAnchor {
 			defFullName = defFullName.Parent()
 		}
-		bsrAnchor = string(defFullName)
+		bsrAnchor := string(defFullName)
 		// For extensions, we use the anchor for the extensions section in the BSR docs.
 		if isExtension {
 			bsrAnchor = "extensions"
 		}
+
 		if bsrHost != "" {
-			docs.WriteString(fmt.Sprintf(
-				"\n[`%s` on the Buf Schema Registry](https://%s/docs/main:%s#%s)",
-				defFullName,
-				bsrHost,
-				s.def.file.ir.Package(),
-				bsrAnchor,
-			))
+			packageName := string(s.def.file.ir.Package())
+			var url string
+			if s.def.file.IsWKT() {
+				// WKT uses special bsrHost format
+				url = "https://" + bsrHost + "/docs/main:" + packageName
+				if bsrAnchor != "" {
+					url += "#" + bsrAnchor
+				}
+			} else {
+				// Use bsrURL for non-WKT modules
+				url = bsrURL(module, packageName, bsrAnchor, bsrTabTypeDocs)
+			}
+			if url != "" {
+				fmt.Fprintf(
+					&docs,
+					"\n[`%s` on the Buf Schema Registry](%s)\n",
+					defFullName,
+					url,
+				)
+			}
 		}
 	}
 	return docs.String()
@@ -679,3 +803,19 @@ func plural(i int) string {
 	}
 	return "s"
 }
+
+// isTrailingComment returns true if the comment has code on the same line before it.
+func isTrailingComment(t token.Token) bool {
+	if t.Kind() != token.Comment {
+		return false
+	}
+	for c := token.NewCursorAt(t); ; {
+		p := c.PrevSkippable()
+		if p.IsZero() || strings.Contains(p.Text(), "\n") {
+			return false
+		}
+		if p.Kind() != token.Space {
+			return true
+		}
+	}
+}
